import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import partial
import logging
from heaserver.service.runner import routes
from heaserver.service import response
from aiohttp import web, hdrs
from heaserver.service.wstl import action, add_run_time_action
from heaserver.service.db import awsservicelib, mongo, aws
from heaserver.service.activity import DesktopObjectActionLifecycle
from heaserver.service.sources import AWS_S3
from heaserver.service.messagebroker import publish_desktop_object
from heaserver.service.heaobjectsupport import new_heaobject_from_type, desktop_object_type_or_type_name_to_type
from heaserver.service.oidcclaimhdrs import SUB
from heaobject.awss3key import encode_key, split, KeyDecodeException, join, replace_parent_folder, decode_key, is_object_in_folder, parent
from heaobject.activity import Status
from heaobject.folder import AWSS3ItemInFolder, AWSS3Folder
from heaobject.project import AWSS3Project
from heaobject.data import AWSS3FileObject, ClipboardData
from heaobject.user import NONE_USER
from heaobject.error import DeserializeException
from heaobject.aws import S3StorageClass
from botocore.exceptions import ClientError as BotoClientError, ParamValidationError
from mypy_boto3_s3 import S3Client
from typing import Any
from datetime import datetime
from json import JSONDecodeError
from yarl import URL

MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION = 'awss3foldersmetadata'

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{folder_id}/items')
@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{folder_id}/items/')
@action(name='heaserver-awss3folders-item-get-actual', rel='hea-actual', path='{+actual_object_uri}')
async def get_items(request: web.Request) -> web.Response:
    """
    Gets the project's items with the specified id.
    :param request: the HTTP request.
    :return: the requested items, or Not Found if the project was not found.
    ---
    summary: All items in a project.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - name: folder_id
          in: path
          required: true
          description: The id of the project to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A folder id
              value: root
    responses:
      '200':
        $ref: '#/components/responses/200'
      '403':
        $ref: '#/components/responses/403'
    """
    logger = logging.getLogger(__name__)

    try:
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        folder_id_ = request.match_info['folder_id']
    except KeyError as e:
        raise ValueError(str(e))
    try:
        folder_key = decode_key(folder_id_)
    except KeyDecodeException:
        return response.status_not_found()

    if not awsservicelib.is_folder(folder_key) or not await _is_project(request, volume_id, bucket_name, folder_id_):
        return response.status_not_found()

    return await _get_items_in_container(request, volume_id, bucket_name, folder_id_)

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{folder_id}/items/{id}')
@action(name='heaserver-awss3folders-item-get-actual', rel='hea-actual',
        path='{+actual_object_uri}')
@action(name='heaserver-awss3folders-item-get-volume', rel='hea-volume', path='/volumes/{volume_id}')
async def get_item(request: web.Request) -> web.Response:
    """
    Gets the requested item from the given project.

    :param request: the HTTP request. Required.
    :return: the requested item, or Not Found if it was not found.
    ---
    summary: A specific project item.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - name: folder_id
          in: path
          required: true
          description: The id of the folder to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A folder id
              value: root
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '403':
        $ref: '#/components/responses/403'
      '404':
        $ref: '#/components/responses/404'
    """
    logger = logging.getLogger(__name__)

    try:
        volume_id = request.match_info['volume_id']
        folder_id_ = request.match_info['folder_id']
        bucket_name = request.match_info['bucket_id']
    except KeyError as e:
        return response.status_bad_request(str(e))
    if 'id' not in request.match_info:
        return response.status_bad_request('id is required')

    decoded_folder_key, decoded_key, folder_or_item_not_found = _check_folder_and_object_keys(folder_id_, request)
    actual_object_uri_base = URL('/volumes') / volume_id / 'buckets' / bucket_name

    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-get',
                                            description=f'Getting item {awsservicelib._activity_object_display_name(bucket_name, decoded_key)}',
                                            activity_cb=publish_desktop_object) as activity:
        async with mongo.MongoContext(request=request, volume_id=volume_id) as mongo_client:
            async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3:
                try:
                    loop = asyncio.get_running_loop()

                    if folder_or_item_not_found:
                        activity.status = Status.FAILED
                        return await awsservicelib.return_bucket_status_or_not_found(bucket_name, loop, s3)

                    logger.debug('Getting item %s in folder %s in bucket %s', decoded_key, decoded_folder_key, bucket_name)
                    response_ = await loop.run_in_executor(None, partial(s3.list_objects_v2, Bucket=bucket_name, Prefix=decoded_key,
                                                                        MaxKeys=1))

                    def _second_to_last(text, pattern):
                        return text.rfind(pattern, 0, text.rfind(pattern))

                    if response_['KeyCount'] > 0:
                        for obj in response_['Contents']:
                            id_ = obj['Key']
                            is_folder_ = awsservicelib.is_folder(id_)
                            id_encoded = encode_key(id_)
                            logger.debug('Found item %s in bucket %s', id_, bucket_name)

                            item = AWSS3ItemInFolder()
                            item.id = id_encoded
                            item.modified = obj['LastModified']
                            item.created = obj['LastModified']
                            item.owner = request.headers.get(SUB, NONE_USER)
                            item.folder_id = folder_id_
                            item.actual_object_id = id_encoded
                            item.storage_class = S3StorageClass[obj['StorageClass']] if not is_folder_ else None
                            item.bucket_id = bucket_name
                            item.source = AWS_S3
                            item.volume_id = volume_id
                            if is_folder_:
                                metadata_dict = await mongo_client.get_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_name, 'encoded_key': id_encoded})
                                if not metadata_dict or metadata_dict['actual_object_type_name'] == AWSS3Folder.get_type_name():
                                    item.actual_object_uri = str(actual_object_uri_base / 'awss3folders' / id_encoded)
                                    item.actual_object_type_name = AWSS3Folder.get_type_name()
                                elif metadata_dict['actual_object_type_name'] == AWSS3Project.get_type_name():
                                    item.actual_object_type_name = AWSS3Project.get_type_name()
                                    item.actual_object_uri = str(actual_object_uri_base / 'awss3projects' / id_encoded)
                            else:
                                item.actual_object_uri = str(actual_object_uri_base / 'awss3files' / id_encoded)
                                item.actual_object_type_name = AWSS3FileObject.get_type_name()
                                item.size = obj['Size']
                            return await response.get(request, item.to_dict())
                    return await response.get(request, None)
                except BotoClientError as e:
                    activity.status = Status.FAILED
                    return awsservicelib.handle_client_error(e)


@routes.route('OPTIONS', '/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
async def get_project_options(request: web.Request) -> web.Response:
    """
    Gets the allowed HTTP methods for a project resource.

    :param request: the HTTP request (required).
    :return: the HTTP response.
    ---
    summary: Allowed HTTP methods.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        description: Expected response to a valid request.
        content:
            text/plain:
                schema:
                    type: string
                    example: "200: OK"
      '404':
        $ref: '#/components/responses/404'
    """
    return await response.get_options(request, ['GET', 'PUT', 'DELETE', 'HEAD', 'OPTIONS'])

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/duplicator')
@action(name='heaserver-awss3folders-project-duplicate-form')
async def get_project_duplicator(request: web.Request) -> web.Response:
    """
    Gets a form template for duplicating the requested project.

    :param request: the HTTP request. Required.
    :return: the requested form, or Not Found if the requested project was not found.
    ---
    summary: A project to duplicate.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project_move_template(request)

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/archive')
@action(name='heaserver-awss3folders-project-archive-form')
async def get_project_archive(request: web.Request) -> web.Response:
    """
    Gets a form template for archiving the requested project.

    :param request: the HTTP request. Required.
    :return: the requested form, or Not Found if the requested project was not found.
    ---
    summary: A specific project to be archived.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '204':
        $ref: '#/components/responses/204'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project_move_template(request)


@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3folders/{id}/mover')
@action(name='heaserver-awss3folders-project-move-form')
async def get_project_mover(request: web.Request) -> web.Response:
    """
    Gets a form template for moving the requested project.

    :param request: the HTTP request. Required.
    :return: the requested form, or Not Found if the requested project was not found.
    ---
    summary: A project to move.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project_move_template(request)


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/archive')
async def post_project_archive(request: web.Request) -> web.Response:
    """
    Posts the provided project to archive it.

    :param request: the HTTP request.
    :return: a Response object with a status of No Content.
    ---
    summary: A specific project to be archived.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    requestBody:
        description: The new name of the project and target for archiving it.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: The name of the storage class to archive the projects to
                  value: {
                    "template": {
                      "data": [
                      {
                        "name": "storage_class",
                        "value": "DEEP_ARCHIVE"
                      }
                      ]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: The new name of the project and target for moving it.
                  value: {
                    "target": "http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/my-bucket/awss3projects/"
                  }
    responses:
      '204':
        $ref: '#/components/responses/204'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        id_ = request.match_info['id']
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        key = awsservicelib.decode_folder(id_)
    except KeyError as e:
        raise ValueError(str(e))
    except KeyDecodeException as e:
        return await response.get(request, None)

    if not awsservicelib.is_folder(key):
        return await response.get(request, None)
    if not await _is_project(request, volume_id, bucket_name, id_):
        return response.status_not_found()
    return await awsservicelib.archive_object(request, activity_cb=publish_desktop_object)


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/unarchive')
async def unarchive_project(request: web.Request) -> web.Response:
    """

    :param request:
    :return: a Response object with status 202 Accept

    ---
    summary: A specific project.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '201':
        $ref: '#/components/responses/201'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        id_ = request.match_info['id']
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        key = awsservicelib.decode_folder(id_)
    except KeyError as e:
        raise ValueError(str(e))
    except KeyDecodeException as e:
        return await response.get(request, None)

    if not awsservicelib.is_folder(key):
        return await response.get(request, None)
    if await _is_project(request, volume_id, bucket_name, id_):
        return await awsservicelib.unarchive_object(request=request, activity_cb=publish_desktop_object)
    else:
        return await response.get(request, None)


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/mover')
async def post_project_mover(request: web.Request) -> web.Response:
    """
    Posts the provided project to move it.

    :param request: the HTTP request.
    :return: a Response object with a status of No Content.
    ---
    summary: A specific project.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    requestBody:
        description: The new name of the project and target for moving it.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: The new name of the project and target for moving it.
                  value: {
                    "template": {
                      "data": [
                      {
                        "name": "target",
                        "value": "http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/my-bucket/awss3projects/"
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: The new name of the project and target for moving it.
                  value: {
                    "target": "http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/my-bucket/awss3projects/"
                  }
    responses:
      '204':
        $ref: '#/components/responses/204'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        id_ = request.match_info.get('id')
    except KeyError as e:
        raise ValueError(str(e))
    try:
        key = awsservicelib.decode_folder(id_)
    except KeyDecodeException as e:
        return response.status_not_found()
    if not awsservicelib.is_folder(key):
        return response.status_not_found()
    async with mongo.MongoContext(request, volume_id) as mongo_client:
        metadata_dict = await mongo_client.get_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_id, 'encoded_key': id_})
        if metadata_dict and metadata_dict['actual_object_type_name'] != AWSS3Project.get_type_name():
            return response.status_not_found()

        try:
            target_uri, target_bucket_id, target_key = await awsservicelib._copy_object_extract_target(await request.json())
        except web.HTTPBadRequest:
            return response.status_bad_request(f'Expected type {AWSS3Project}; actual object was {await request.text()}')

        async with DesktopObjectActionLifecycle(request=request,
                                                code='hea-move',
                                                description=f'Moving {awsservicelib._activity_object_display_name(bucket_id, key)} to {awsservicelib._activity_object_display_name(target_bucket_id, target_key)}',
                                                activity_cb=publish_desktop_object) as activity:
            async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
                try:
                    target_key_ = join(target_key, split(key)[1])
                    await _move(s3_client=s3_client, source_bucket_id=bucket_id, source_key=key, target_bucket_id=target_bucket_id, target_key=target_key_)
                    if metadata_dict:
                        metadata_dict['parent_encoded_key'] = encode_key(target_key)
                        metadata_dict['encoded_key'] = encode_key(target_key_)
                        await mongo_client.update_admin(metadata_dict, MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION)
                    activity.new_object_uri = target_uri
                    return response.status_no_content()
                except BotoClientError as e:
                    activity.status = Status.FAILED
                    return awsservicelib.handle_client_error(e)



@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/duplicator')
async def post_project_duplicator(request: web.Request) -> web.Response:
    """
    Posts the provided project for duplication.

    :param request: the HTTP request.
    :return: a Response object with a status of Created and the object's URI in the
    ---
    summary: A specific folder item.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    requestBody:
        description: The target folder or project.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: The target folder or project.
                  value: {
                    "template": {
                      "data": [
                      {
                        "name": "target",
                        "value": "http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/my-bucket"
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: The target folder or project.
                  value: {
                    "target": "http://localhost:8080/volumes/666f6f2d6261722d71757578/buckets/my-bucket"
                  }
    responses:
      '201':
        $ref: '#/components/responses/201'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        id_ = request.match_info['id']
    except KeyError as e:
        raise ValueError(str(e))
    try:
        key = awsservicelib.decode_folder(id_)
    except KeyDecodeException as e:
        return response.status_not_found()
    if not awsservicelib.is_folder(key):
        return response.status_not_found()
    if not await _is_project(request, volume_id, bucket_id, id_):
        return response.status_not_found()
    return await awsservicelib.copy_object(request, activity_cb=publish_desktop_object)


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/uploader')
@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/uploader/')
async def post_project_uploader(request: web.Request) -> web.Response:
    """
    Uploads file to a target project.

    :param request: the HTTP request. The body of the request is expected to be an item or an actual object.
    :return: the response, with a 204 status code if an item was created or a 400 if not. If an item was created, the
    Location header will contain the URL of the created item.
    ---
    summary: A upload to a specific folder item.
    tags:
        - heaserver-folders-project-uploader
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - name: id
          in: path
          required: true
          description: The id of the folder.
          schema:
            type: string
          examples:
            example:
              summary: A folder id
              value: root
    requestBody:
        description: A new folder item object.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: Item example
                  value: {
                    "template": {
                      "data": [{
                        "name": "created",
                        "value": null
                      },
                      {
                        "name": "derived_by",
                        "value": null
                      },
                      {
                        "name": "derived_from",
                        "value": []
                      },
                      {
                        "name": "description",
                        "value": null
                      },
                      {
                        "name": "display_name",
                        "value": "Bob"
                      },
                      {
                        "name": "invited",
                        "value": []
                      },
                      {
                        "name": "modified",
                        "value": null
                      },
                      {
                        "name": "name",
                        "value": "bob"
                      },
                      {
                        "name": "owner",
                        "value": "system|none"
                      },
                      {
                        "name": "shares",
                        "value": []
                      },
                      {
                        "name": "source",
                        "value": null
                      },
                      {
                        "name": "version",
                        "value": null
                      },
                      {
                        "name": "actual_object_id",
                        "value": "666f6f2d6261722d71757578"
                      },
                      {
                        "name": "actual_object_type_name",
                        "value": "heaobject.data.AWSS3FileObject"
                      },
                      {
                        "name": "actual_object_uri",
                        "value": "/volumes/666f6f2d6261722d71757578/buckets/my-bucket/folders/666f6f2d6261722d71757578"
                      },
                      {
                        "name": "type",
                        "value": "heaobject.folder.AWSS3ItemInFolder"
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: Item example
                  value: {
                    "created": null,
                    "derived_by": null,
                    "derived_from": [],
                    "description": null,
                    "display_name": "Joe",
                    "invited": [],
                    "modified": null,
                    "name": "joe",
                    "owner": "system|none",
                    "shares": [],
                    "source": null,
                    "type": "heaobject.folder.AWSS3Item",
                    "version": null,
                    "folder_id": "root",
                    "actual_object_id": "666f6f2d6261722d71757578",
                    "actual_object_type_name": "heaobject.registry.Component",
                    "actual_object_uri": "/volumes/666f6f2d6261722d71757578/buckets/my-bucket/folders/666f6f2d6261722d71757578"
                  }
    responses:
      '201':
        $ref: '#/components/responses/201'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    logger = logging.getLogger(__name__)
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        folder_id = request.match_info.get('folder_id') if request.match_info.get('folder_id') else request.match_info.get('id')
    except KeyError as e:
        return response.status_bad_request(str(e))

    try:
        obj = await new_heaobject_from_type(request, AWSS3FileObject)
    except DeserializeException as e:
        return response.status_bad_request(str(e).encode())
    if obj is None:
        return response.status_bad_request('The request body must have a desktop object')
    if obj.display_name is None:
        return response.status_bad_request("display_name is required")
    if '/' in obj.display_name:
        return response.status_bad_request(f"The item's display name may not have slashes in it")
    try:
        # if folder id is root let it decode to '' otherwise if its not a Folder set it to None
        target_folder_id = awsservicelib.decode_folder(folder_id)
        if folder_id != awsservicelib.ROOT_FOLDER.id and not awsservicelib.is_folder(target_folder_id):
            target_folder_id = None
    except KeyDecodeException:
        target_folder_id = None
    if not await _is_project(request, volume_id, bucket_id, folder_id):
        return response.status_not_found()
    logger.debug(f"the file obj's name and display_name {obj.name}, {obj.display_name}")


    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-create',
                                            description=f'Creating item {obj.display_name} in bucket {bucket_id}',
                                            activity_cb=publish_desktop_object) as activity:
        async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
            loop = asyncio.get_running_loop()
            try:
                if target_folder_id is None:
                    await loop.run_in_executor(None, partial(s3_client.head_bucket, Bucket=bucket_id))
                    return response.status_not_found()
                obj_name = f'{target_folder_id}{obj.display_name}'
                url = f"volumes/{request.match_info['volume_id']}/buckets/{request.match_info['bucket_id']}/awss3files"
                content_id = f"{encode_key(obj_name)}/content"
                # check if item exists, if not throws an exception
                await loop.run_in_executor(None, partial(s3_client.head_object, Bucket=bucket_id, Key=obj_name))
                return response.status_ok(headers={hdrs.LOCATION: f"/{url}/{content_id}"})
            except BotoClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == awsservicelib.CLIENT_ERROR_404:  # file metadata doesn't exist
                    await loop.run_in_executor(None, partial(s3_client.put_object, Bucket=bucket_id, Key=obj_name))
                    return await response.post(request, f"{encode_key(obj_name)}/content", url)
                elif error_code == awsservicelib.CLIENT_ERROR_NO_SUCH_BUCKET:
                    activity.status = Status.FAILED
                    return response.status_not_found()
                else:
                    activity.status = Status.FAILED
                    return response.status_bad_request(str(e))


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{folder_id}/items')
@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{folder_id}/items/')
async def post_item_in_project(request: web.Request) -> web.Response:
    """
    Creates a new item in a project.

    :param request: the HTTP request. The body of the request is expected to be an item or an actual object.
    :return: the response, with a 204 status code if an item was created or a 400 if not. If an item was created, the
    Location header will contain the URL of the created item.
    ---
    summary: A specific project item.
    tags:
        - heaserver-folders-project-items
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - name: folder_id
          in: path
          required: true
          description: The id of the folder.
          schema:
            type: string
          examples:
            example:
              summary: A folder id
              value: root
    requestBody:
        description: A new project item object.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: Item example
                  value: {
                    "template": {
                      "data": [{
                        "name": "created",
                        "value": null
                      },
                      {
                        "name": "derived_by",
                        "value": null
                      },
                      {
                        "name": "derived_from",
                        "value": []
                      },
                      {
                        "name": "description",
                        "value": null
                      },
                      {
                        "name": "display_name",
                        "value": "Bob"
                      },
                      {
                        "name": "invited",
                        "value": []
                      },
                      {
                        "name": "modified",
                        "value": null
                      },
                      {
                        "name": "name",
                        "value": "bob"
                      },
                      {
                        "name": "owner",
                        "value": "system|none"
                      },
                      {
                        "name": "shares",
                        "value": []
                      },
                      {
                        "name": "source",
                        "value": null
                      },
                      {
                        "name": "version",
                        "value": null
                      },
                      {
                        "name": "actual_object_id",
                        "value": "666f6f2d6261722d71757578"
                      },
                      {
                        "name": "actual_object_type_name",
                        "value": "heaobject.data.AWSS3FileObject"
                      },
                      {
                        "name": "actual_object_uri",
                        "value": "/volumes/666f6f2d6261722d71757578/buckets/my-bucket/folders/666f6f2d6261722d71757578"
                      },
                      {
                        "name": "type",
                        "value": "heaobject.folder.AWSS3ItemInFolder"
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: Item example
                  value: {
                    "created": null,
                    "derived_by": null,
                    "derived_from": [],
                    "description": null,
                    "display_name": "Joe",
                    "invited": [],
                    "modified": null,
                    "name": "joe",
                    "owner": "system|none",
                    "shares": [],
                    "source": null,
                    "type": "heaobject.folder.AWSS3Item",
                    "version": null,
                    "folder_id": "root",
                    "actual_object_id": "666f6f2d6261722d71757578",
                    "actual_object_type_name": "heaobject.registry.Component",
                    "actual_object_uri": "/volumes/666f6f2d6261722d71757578/buckets/my-bucket/folders/666f6f2d6261722d71757578"
                  }
    responses:
      '201':
        $ref: '#/components/responses/201'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        item = await new_heaobject_from_type(request, AWSS3ItemInFolder)
    except DeserializeException as e:
        return response.status_bad_request(str(e).encode())

    logger = logging.getLogger(__name__)
    if item.display_name is None:
        return response.status_bad_request("display_name is required")
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        folder_id = request.match_info['folder_id']
    except KeyError as e:
        raise ValueError(str(e))

    if item is None:
        return response.status_bad_request('item is a required field')

    if item.folder_id is not None and folder_id != item.folder_id:
        return response.status_bad_request(
            f'folder_id in object was {item.folder_id} but folder_id in URL was {folder_id}')
    if item.folder_id is not None and item.folder_id != folder_id:
        return response.status_bad_request(
            f'Inconsistent folder_id in URL versus item: {folder_id} vs {item.folder_id}')
    if '/' in item.display_name:
        return response.status_bad_request(f"The item's display name may not have slashes in it")
    try:
        # if folder id is root let it decode to '' otherwise if its not a Folder set it to None
        folder_key = awsservicelib.decode_folder(folder_id)

        if folder_id != awsservicelib.ROOT_FOLDER.id and not awsservicelib.is_folder(folder_key):
            folder_key = None
    except KeyDecodeException:
        folder_key = None

    if not await _is_project(request, volume_id, bucket_id, folder_key):
        return response.status_not_found()

    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-create',
                                            description=f'Creating item {item.display_name} in bucket {bucket_id}',
                                            activity_cb=publish_desktop_object) as activity:
        async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
            loop = asyncio.get_running_loop()
            try:
                if folder_key is None:
                    await loop.run_in_executor(None, partial(s3_client.head_bucket, Bucket=bucket_id))
                    activity.status = Status.FAILED
                    return response.status_not_found()
                item_name = f'{folder_key}{item.display_name}'
                if item.actual_object_type_name in (AWSS3Folder.get_type_name(), AWSS3Project.get_type_name()):
                    item_name += '/'
                elif item.actual_object_type_name == AWSS3FileObject.get_type_name():
                    pass
                else:
                    activity.status = Status.FAILED
                    return response.status_bad_request(f'Unsupported actual_object_type_name {item.actual_object_type_name}')
                response_ = await loop.run_in_executor(None, partial(s3_client.head_object, Bucket=bucket_id,
                                                                    Key=item_name))  # check if item exists, if not throws an exception
                logger.debug('Result of post_object: %s', response_)
                activity.status = Status.FAILED
                return response.status_bad_request(body=f"Item {item_name} already exists")
            except BotoClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == awsservicelib.CLIENT_ERROR_404:  # folder doesn't exist
                    await loop.run_in_executor(None, partial(s3_client.put_object, Bucket=bucket_id, Key=item_name))
                    logger.info('Added item %s', item_name)
                    if item.actual_object_type_name == AWSS3Project.get_type_name():
                        async with mongo.MongoContext(request, volume_id) as mongo_client:
                            metadata = {
                                'bucket_id': bucket_id,
                                'encoded_key': item.id,
                                'actual_object_type_name': AWSS3Project.get_type_name(),
                                'parent_encoded_key': folder_id
                            }
                            await mongo_client.upsert_admin(metadata, MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION)
                    return await response.post(request, encode_key(item_name),
                                            f"volumes/{request.match_info['volume_id']}/buckets/{request.match_info['bucket_id']}/awss3projects/{folder_id}/items")
                elif error_code == awsservicelib.CLIENT_ERROR_NO_SUCH_BUCKET:
                    activity.status = Status.FAILED
                    return response.status_not_found()
                else:
                    activity.status = Status.FAILED
                    return response.status_bad_request(str(e))

@routes.delete('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{folder_id}/items/{id}')
async def delete_item(request: web.Request) -> web.Response:
    """
    Deletes the item with the specified id and its corresponding desktop object.

    :param request: the HTTP request.
    :return: No Content or Not Found.
    ---
    summary: Folder item deletion
    tags:
        - heaserver-folders-folder-items
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - name: folder_id
          in: path
          required: true
          description: The id of the project.
          schema:
            type: string
          examples:
            example:
              summary: A project id
              value: root
        - $ref: '#/components/parameters/id'
    responses:
      '204':
        $ref: '#/components/responses/204'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        id_ = request.match_info['id']
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        key = awsservicelib.decode_folder(id_)
    except KeyError as e:
        raise ValueError(str(e))
    except KeyDecodeException:
        return response.status_not_found()
    if not awsservicelib.is_folder(key) or not await _is_project(request, volume_id, bucket_id, key):
        return response.status_not_found()
    response_ = await awsservicelib.delete_object(request, recursive=True, activity_cb=publish_desktop_object)
    if response_.status == 204:
        async with mongo.MongoContext(request, volume_id) as mongo_client:
            await mongo_client.delete_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_id, 'encoded_key': id_})
    return response_

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
@action(name='heaserver-awss3folders-project-get-open-choices', rel='hea-opener-choices hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/opener')
@action(name='heaserver-awss3folders-project-get-properties', rel='hea-properties hea-context-menu')
@action(name='heaserver-awss3folders-project-duplicate', rel='hea-dynamic-standard hea-icon-duplicator hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/duplicator')
@action(name='heaserver-awss3folders-project-move', rel='hea-dynamic-standard hea-icon-mover hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/mover')
@action(name='heaserver-awss3folders-project-get-create-choices', rel='hea-creator-choices hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/creator')
@action(name='heaserver-awss3folders-project-get-trash', rel='hea-trash hea-context-menu',
        path='/volumes/{volume_id}/awss3trash')
@action(name='heaserver-awss3folders-project-archive', rel='hea-archive hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/archive')
@action(name='heaserver-awss3folders-project-unarchive', rel='hea-unarchive hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/unarchive')
@action(name='heaserver-awss3folders-project-get-presigned-url',
        rel='hea-dynamic-clipboard hea-icon-for-clipboard hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/presignedurl')
@action(name='heaserver-awss3folders-project-get-self', rel='self',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
@action(name='heaserver-awss3folders-project-get-volume', rel='hea-volume', path='/volumes/{volume_id}')
@action(name='heaserver-awss3folders-project-get-awsaccount', rel='hea-account',
        path='/volumes/{volume_id}/awsaccounts/me')
@action(name='heaserver-awss3folders-item-upload', rel='hea-uploader',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/uploader')
async def get_project(request: web.Request) -> web.Response:
    """
    Gets the project with the specified id.

    :param request: the HTTP request.
    :return: the requested project or Not Found.
    ---
    summary: A specific folder.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project(request)


@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/byname/{name}')
@action(name='heaserver-awss3folders-project-get-self', rel='self',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
@action(name='heaserver-awss3folders-project-get-volume', rel='hea-volume', path='/volumes/{volume_id}')
@action(name='heaserver-awss3folders-project-get-awsaccount', rel='hea-account',
        path='/volumes/{volume_id}/awsaccounts/me')
async def get_project_by_name(request: web.Request) -> web.Response:
    """
    Gets the project with the specified name.

    :param request: the HTTP request.
    :return: the requested project or Not Found.
    ---
    summary: A specific project.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/name'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project(request)

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/presignedurl')
@action(name='heaserver-awss3folders-project-get-presigned-url-form')
async def get_presigned_url_form(request: web.Request) -> web.Response:
    """
    Returns a template for requesting the generation of a presigned URL.

    :param request: the HTTP request.
    :return: No Content or Not Found.
    ---
    summary: Presigned url for folders
    tags:
        - heaserver-folders-folders
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '403':
        $ref: '#/components/responses/403'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project(request)

@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/presignedurl')
async def post_presigned_url_form(request: web.Request) -> web.Response:
    """
    Posts a template for requesting the generation of a presigned URL.

    :param request: the HTTP request.
    :return: No Content or Not Found.
    ---
    summary: Presigned url for projects
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    requestBody:
        description: The expiration time for the presigned URL.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: The expiration time for the presigned URL.
                  value: {
                    "template": {
                      "data": [
                      {
                        "name": "link_expiration",
                        "value": 259200
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: The new name of the file and target for moving it.
                  value: {
                    "link_expiration": 259200
                  }
    responses:
      '200':
        $ref: '#/components/responses/200'
      '403':
        $ref: '#/components/responses/403'
      '404':
        $ref: '#/components/responses/404'
    """
    logger = logging.getLogger(__name__)
    # Generate a presigned URL for the S3 object
    volume_id = request.match_info['volume_id']
    bucket_id = request.match_info['bucket_id']
    object_id = request.match_info['id']
    # three days default for expiration
    try:
        expiration_seconds = await _extract_expiration(await request.json())
    except JSONDecodeError as e:
        return response.status_bad_request(str(e))

    try:
        object_key = awsservicelib.decode_folder(object_id)
    except KeyDecodeException:
        # Let the bucket query happen so that we consistently return Forbidden if the user lacks permissions
        # for the bucket.
        return response.status_not_found()
    if not await _is_project(request, volume_id, bucket_id, object_id):
        return response.status_not_found()

    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-get',
                                            description=f'Generating presigned-URL(s) for {awsservicelib._activity_object_display_name(bucket_id, object_key)}',
                                            activity_cb=publish_desktop_object) as activity:
        async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
            try:
                loop = asyncio.get_running_loop()
                urls = []
                async for obj in awsservicelib.list_object_versions(s3_client, bucket_id, object_key, loop):
                    url = await loop.run_in_executor(None, partial(s3_client.generate_presigned_url, 'get_object',
                                                                Params={'Bucket': bucket_id, 'Key': obj['Key']},
                                                                ExpiresIn=expiration_seconds if expiration_seconds is not None else 259200))
                    urls.append(url)
            except BotoClientError as e:
                activity.status = Status.FAILED
                return awsservicelib.handle_client_error(e)
            # The response contains the presigned URL
            data = ClipboardData()
            data.mime_type = 'text/plain;encoding=utf-8'
            data.data = _client_line_ending(request).join(urls)
            data.created = datetime.now().astimezone()
            f = AWSS3FileObject()
            f.bucket_id = bucket_id
            f.id = object_id
            data.display_name = f'Presigned URL for {f.display_name}'
            return await response.get(request, data.to_dict())

@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects')
@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/')
async def post_project(request: web.Request) -> web.Response:
    """
    Creates a new project.

    :param request: the HTTP request. The body of the request is expected to be a project.
    :return: the response, with a 204 status code if a project was created. If a
    folder was created, the Location header will contain the URL of the created
    project.
    ---
    summary: A specific project.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
    requestBody:
        description: A new project object.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: Project example
                  value: {
                    "template": {
                      "data": [{
                        "name": "created",
                        "value": null
                      },
                      {
                        "name": "derived_by",
                        "value": null
                      },
                      {
                        "name": "derived_from",
                        "value": []
                      },
                      {
                        "name": "description",
                        "value": null
                      },
                      {
                        "name": "display_name",
                        "value": "Bob"
                      },
                      {
                        "name": "invited",
                        "value": []
                      },
                      {
                        "name": "modified",
                        "value": null
                      },
                      {
                        "name": "name",
                        "value": "bob"
                      },
                      {
                        "name": "owner",
                        "value": "system|none"
                      },
                      {
                        "name": "shares",
                        "value": []
                      },
                      {
                        "name": "source",
                        "value": null
                      },
                      {
                        "name": "version",
                        "value": null
                      },
                      {
                        "name": "type",
                        "value": "heaobject.project.AWSS3Project"
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: Project example
                  value: {
                    "created": null,
                    "derived_by": null,
                    "derived_from": [],
                    "description": null,
                    "display_name": "Joe",
                    "invited": [],
                    "modified": null,
                    "name": "joe",
                    "owner": "system|none",
                    "shares": [],
                    "source": null,
                    "type": "heaobject.project.AWSS3Project",
                    "version": null
                  }
    responses:
      '201':
        $ref: '#/components/responses/201'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
    except KeyError as e:
        return response.status_bad_request(str(e))

    try:
        project = await new_heaobject_from_type(request, AWSS3Project)
    except TypeError:
        return response.status_bad_request(f'Expected type {AWSS3Project}; actual object was {await request.text()}')
    except DeserializeException as e:
        return response.status_bad_request(str(e))
    response_ = await awsservicelib.create_object(request, AWSS3Project)

    if response_.status == 201:
        async with mongo.MongoContext(request, volume_id) as mongo_client:
            metadata = {
                'bucket_id': bucket_id,
                'encoded_key': project.id,
                'actual_object_type_name': AWSS3Project.get_type_name(),
                'parent_encoded_key': parent(project.key)
            }
            await mongo_client.upsert_admin(metadata,
                                            MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION,
                                            mongoattributes={'bucket_id': bucket_id, 'encoded_key': project.id})
    return response_

@routes.put('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
async def put_project(request: web.Request) -> web.Response:
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        id_ = request.match_info.get('id')
    except KeyError as e:
        raise ValueError(str(e))

    try:
        key = awsservicelib.decode_folder(id_)
    except KeyDecodeException as e:
        return response.status_not_found()
    if not awsservicelib.is_folder(key) or not await _is_project(request, volume_id, bucket_id, id_):
        return response.status_not_found()

    async with mongo.MongoContext(request, volume_id) as mongo_client:
        metadata_dict = await mongo_client.get_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_id, 'encoded_key': id_})
        if metadata_dict and metadata_dict['actual_object_type_name'] != AWSS3Project.get_type_name():
            return response.status_not_found()
        try:
            project = await new_heaobject_from_type(request, AWSS3Project)
        except TypeError:
            return response.status_bad_request(f'Expected type {AWSS3Project}; actual object was {await request.text()}')
        except DeserializeException as e:
            return response.status_bad_request(str(e))
        if key != project.key:
            if bucket_id != project.bucket_id:
                return response.status_bad_request(f"The project's bucket id was {project.bucket_id} but the URL's bucket id was {bucket_id}")

            async with DesktopObjectActionLifecycle(request,
                                                    code='hea-update',
                                                    description=f'Renaming {awsservicelib._activity_object_display_name(bucket_id, key)} to {project.key}',
                                                    activity_cb=publish_desktop_object) as activity:
                async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
                    try:
                        await _move(s3_client=s3_client, source_bucket_id=bucket_id, source_key=key, target_bucket_id=bucket_id, target_key=project.key)
                        if metadata_dict:
                            metadata_dict['parent_encoded_key'] = encode_key(project.key)
                            metadata_dict['encoded_key'] = encode_key(join(project.key, split(key)[1]))
                            await mongo_client.update_admin(metadata_dict, MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION)
                    except BotoClientError as e:
                        activity.status = Status.FAILED
                        return awsservicelib.handle_client_error(e)

        return await response.put(True)

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects')
@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/')
@action('heaserver-awss3folders-project-get-open-choices', rel='hea-opener-choices hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/opener')
@action(name='heaserver-awss3folders-project-get-properties', rel='hea-properties hea-context-menu')
@action(name='heaserver-awss3folders-project-duplicate', rel='hea-dynamic-standard hea-icon-duplicator hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/duplicator')
@action(name='heaserver-awss3folders-project-move', rel='hea-dynamic-standard hea-icon-mover hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/mover')
@action(name='heaserver-awss3folders-project-get-create-choices', rel='hea-creator-choices hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/creator')
@action(name='heaserver-awss3folders-project-archive', rel='hea-archive hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/archive')
@action(name='heaserver-awss3folders-project-unarchive', rel='hea-unarchive hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/unarchive')
@action(name='heaserver-awss3folders-project-get-trash', rel='hea-trash hea-context-menu',
        path='/volumes/{volume_id}/awss3trash')
@action(name='heaserver-awss3folders-project-get-presigned-url',
        rel='hea-dynamic-clipboard hea-icon-for-clipboard hea-context-menu',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/presignedurl')
@action(name='heaserver-awss3folders-project-get-self', rel='self',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
@action(name='heaserver-awss3folders-item-upload', rel='hea-uploader',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/uploader')
async def get_projects(request: web.Request) -> web.Response:
    """
    Gets all projects in a bucket.

    :param request: the HTTP request.
    :return: the requested projects or Not Found.
    ---
    summary: Gets all projects in a bucket.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    logger = logging.getLogger(__name__)
    try:
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
    except KeyError as e:
        raise ValueError(str(e))

    async with mongo.MongoContext(request, volume_id) as mongo_client:
        async with DesktopObjectActionLifecycle(request=request,
                                                code='hea-get',
                                                description=f'Listing all projects in bucket {bucket_name}',
                                                activity_cb=publish_desktop_object) as activity:
            async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3:
                try:
                    loop = asyncio.get_running_loop()
                    logger.debug('Getting all projects from bucket %s', bucket_name)
                    folder_dicts = []
                    async for obj in awsservicelib.list_objects(s3, bucket_name, loop=loop):
                        logger.debug('Checking possible project %s in bucket %s', obj, bucket_name)
                        key = obj['Key']
                        if awsservicelib.is_folder(key):
                            encoded_key = encode_key(key)
                            logger.debug('Project %s in bucket %s is a folder', key, bucket_name)
                            folder_dict = _new_project(bucket_name, obj, encoded_key, request)
                            folder_dicts.append(folder_dict.to_dict())
                    folder_metadata = {}
                    async for metadata_dict in mongo_client.get_all_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_name}):
                        folder_metadata[metadata_dict['encoded_key']] = metadata_dict
                    real_folder_dicts = []
                    for folder_dict in folder_dicts:
                        if (obj_metadata_ := folder_metadata.get(folder_dict['id'])) is None or obj_metadata_['actual_object_type_name'] == AWSS3Project.get_type_name():
                            real_folder_dicts.append(folder_dict)
                    return await response.get_all(request, real_folder_dicts)
                except BotoClientError as e:
                    activity.status = Status.FAILED
                    return awsservicelib.handle_client_error(e)
                except ParamValidationError as e:
                    activity.status = Status.FAILED
                    return response.status_not_found()

@routes.route('OPTIONS', '/volumes/{volume_id}/buckets/{bucket_id}/awss3projects')
@routes.route('OPTIONS', '/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/')
async def get_projects_options(request: web.Request) -> web.Response:
    """
    Gets the allowed HTTP methods for the projects resource.

    :param request: the HTTP request (required).
    :response: the HTTP response.
    ---
    summary: Allowed HTTP methods.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
    responses:
      '200':
        description: Expected response to a valid request.
        content:
            text/plain:
                schema:
                    type: string
                    example: "200: OK"
      '403':
        $ref: '#/components/responses/403'
      '404':
        $ref: '#/components/responses/404'
    """
    return await response.get_options(request, ['GET', 'DELETE', 'HEAD', 'OPTIONS'])

@routes.delete('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}')
async def delete_project(request: web.Request) -> web.Response:
    """
    Deletes the project with the specified id.

    :param request: the HTTP request.
    :return: No Content or Not Found.
    ---
    summary: Deletes the project with the specified id.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '204':
        $ref: '#/components/responses/204'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        id_ = request.match_info['id']
        key = awsservicelib.decode_folder(id_)
    except KeyError as e:
        raise ValueError(str(e))
    except KeyDecodeException:
        return response.status_not_found()
    if not awsservicelib.is_folder(key) or not await _is_project(request, volume_id, bucket_id, id_):
        return response.status_not_found()
    response_ = await awsservicelib.delete_object(request, recursive=True, activity_cb=publish_desktop_object)
    if response_.status == 204:
        async with mongo.MongoContext(request, volume_id) as mongo_client:
            await mongo_client.delete_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_id, 'encoded_key': id_})
    return response_

@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/newfolder')
@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/newfolder/')
async def post_new_folder(request: web.Request) -> web.Response:
    """
    Creates a new folder within a project.

    :param request: the HTTP request. Required.
    :return: Status code 201 if successful, with the URL to the new folder in
    the Location header.
    ---
    summary: Creates a new folder within a project.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    requestBody:
        description: A new folder.
        required: true
        content:
            application/vnd.collection+json:
              schema:
                type: object
              examples:
                example:
                  summary: Folder example
                  value: {
                    "template": {
                      "data": [
                      {
                        "name": "display_name",
                        "value": "Bob"
                      },
                      {
                        "name": "type",
                        "value": "heaobject.folder.AWSS3Folder"
                      }]
                    }
                  }
            application/json:
              schema:
                type: object
              examples:
                example:
                  summary: Item example
                  value: {
                    "display_name": "Joe",
                    "type": "heaobject.folder.AWSS3Folder"
                  }
    responses:
      '201':
        $ref: '#/components/responses/201'
      '400':
        $ref: '#/components/responses/400'
      '404':
        $ref: '#/components/responses/404'
    """
    logger = logging.getLogger(__name__)

    try:
        folder = await new_heaobject_from_type(request, AWSS3Folder)
    except DeserializeException as e:
        return response.status_bad_request(f'Invalid new folder: {str(e)}')
    if folder is None:
        return response.status_bad_request('The request body must contain a folder')
    if folder.display_name is None:
        return response.status_bad_request("display_name is required")
    if '/' in folder.display_name:
        return response.status_bad_request("The folder's display name may not have slashes in it")
    try:
        volume_id = request.match_info['volume_id']
        bucket_id = request.match_info['bucket_id']
        id_ = request.match_info['id']
    except KeyError as e:
        return response.status_bad_request(str(e))

    if id_ == awsservicelib.ROOT_FOLDER.id:
        item_folder_id = ''
    else:
        try:
            item_folder_id = awsservicelib.decode_folder(id_)
            if not awsservicelib.is_folder(item_folder_id) or not _is_project(request, volume_id, bucket_id, item_folder_id):
                item_folder_id = None
        except KeyDecodeException:
            item_folder_id = None
    item_name = join(item_folder_id, folder.display_name)

    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-create',
                                            description=f'Creating {awsservicelib._activity_object_display_name(bucket_id, folder.key)}',
                                            activity_cb=publish_desktop_object) as activity:
        async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
            loop = asyncio.get_running_loop()
            try:
                if item_folder_id is None:
                    await loop.run_in_executor(None, partial(s3_client.head_bucket, Bucket=bucket_id))
                    activity.status = Status.FAILED
                    return response.status_not_found()
                if folder.type == AWSS3Folder.get_type_name():
                    item_name += '/'
                else:
                    activity.status = Status.FAILED
                    return response.status_bad_request(f'Unsupported type {folder.get_type_name()}')
                response_ = await loop.run_in_executor(None, partial(s3_client.head_object, Bucket=bucket_id,
                                                                    Key=item_name))  # check if folder exists, if not throws an exception
                logger.debug('Result of post_object: %s', response_)
                activity.status = Status.FAILED
                return response.status_bad_request(body=f"Item {item_name} already exists")
            except BotoClientError as e:
                error_code = e.response['Error']['Code']
                if error_code == awsservicelib.CLIENT_ERROR_404:  # folder doesn't exist
                    await loop.run_in_executor(None, partial(s3_client.put_object, Bucket=bucket_id, Key=item_name))
                    logger.debug('Added folder %s', item_name)
                    return await response.post(request, encode_key(item_name),
                                            f"volumes/{volume_id}/buckets/{bucket_id}/awss3folders")
                elif error_code == awsservicelib.CLIENT_ERROR_NO_SUCH_BUCKET:
                    activity.status = Status.FAILED
                    return response.status_not_found()
                else:
                    activity.status = Status.FAILED
                    return response.status_bad_request(str(e))

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/uploader')
@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/uploader/')
@action('heaserver-awss3folders-project-upload-form')
async def get_project_uploader_form(request: web.Request) -> web.Response:
    """
    Gets a form template for uploading a file to a project

    :param request: the HTTP request. Required.
    :return: a blank form for uploading a file to a project or Not Found if the r
    equested project does not exist.
    """
    return await _get_project(request)

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/newfolder')
@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/newfolder/')
@action('heaserver-awss3folders-folder-new-form')
async def get_new_folder_form(request: web.Request) -> web.Response:
    """
    Gets form template for creating a new folder within a project.

    :param request: the HTTP request. Required.
    :return: a template for creating a folder or Not Found if the requested project does not exist.
    ---
    summary: A form template for creating a new folder within a projecct.
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await _get_project(request)

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/creator')
@action('heaserver-awss3folders-project-create-folder', rel='hea-creator hea-default application/x.folder',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/newfolder')
async def get_project_creator(request: web.Request) -> web.Response:
    """
    Opens the requested folder.

    :param request: the HTTP request. Required.
    :return: Opener choices, or Not Found if the requested project does not exist.
    ---
    summary: Project creator choices
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '300':
        $ref: '#/components/responses/300'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        folder_name = request.match_info['id'] if 'id' in request.match_info else request.match_info['name']
    except KeyError as e:
        raise ValueError(str(e))
    try:
        folder_key = awsservicelib.decode_folder(folder_name)
    except KeyDecodeException as e:
        return response.status_not_found()

    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-get',
                                            description=f'Accessing {awsservicelib._activity_object_display_name(bucket_name, folder_key)}',
                                            activity_cb=publish_desktop_object) as activity:
        async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
            try:
                folder = await _get_project_helper(s3_client, request)
                if not await _is_project(request, volume_id, bucket_name, folder_name):
                    return response.status_not_found()
                return await response.get_multiple_choices(request, folder.to_dict())
            except BotoClientError as e:
                activity.status = Status.FAILED
                return awsservicelib.handle_client_error(e)
            except web.HTTPClientError as e:
                activity.status = Status.FAILED
                return e

@routes.get('/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/opener')
@action('heaserver-awss3folders-project-open-default', rel='hea-opener hea-default application/x.project',
        path='/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/{id}/items/')
async def get_project_opener(request: web.Request) -> web.Response:
    """
    Opens the requested folder.

    :param request: the HTTP request. Required.
    :return: Opener choices, or Not Found if the requested project does not exist.
    ---
    summary: Project opener choices
    tags:
        - heaserver-folders-projects
    parameters:
        - name: volume_id
          in: path
          required: true
          description: The id of the volume to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A volume id
              value: 666f6f2d6261722d71757578
        - name: bucket_id
          in: path
          required: true
          description: The id of the bucket to retrieve.
          schema:
            type: string
          examples:
            example:
              summary: A bucket id
              value: my-bucket
        - $ref: '#/components/parameters/id'
    responses:
      '300':
        $ref: '#/components/responses/300'
      '404':
        $ref: '#/components/responses/404'
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        folder_name = request.match_info['id'] if 'id' in request.match_info else request.match_info['name']
    except KeyError as e:
        return response.status_bad_request(str(e))
    try:
        folder_key = awsservicelib.decode_folder(folder_name)
    except KeyDecodeException as e:
        return response.status_not_found()

    async with DesktopObjectActionLifecycle(request=request,
                                            code='hea-get',
                                            description=f'Accessing {awsservicelib._activity_object_display_name(bucket_name, folder_key)}',
                                            activity_cb=publish_desktop_object) as activity:
        async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
            try:
                folder = await _get_project_helper(s3_client, request)
                return await response.get_multiple_choices(request, folder.to_dict())
            except BotoClientError as e:
                activity.status = Status.FAILED
                return awsservicelib.handle_client_error(e)
            except web.HTTPClientError as e:
                activity.status = Status.FAILED
                return e

def _client_line_ending(request: web.Request) -> str:
    user_agent = request.headers.get(hdrs.USER_AGENT)
    return '\r\n' if 'Windows' in user_agent else '\n'

async def _extract_expiration(body: dict[str, Any]) -> tuple[str, int]:
    """
    Extracts the target URL and expiration time for a presigned URL request. It un-escapes them
    as needed.

    :param body: a Collection+JSON template dict.
    :return: a three-tuple containing the target URL and the un-escaped expiration time in seconds.
    :raises web.HTTPBadRequest: if the given body is invalid.
    """
    try:
        expiration_seconds = next(
            int(item['value']) for item in body['template']['data'] if item['name'] == 'link_expiration')
        return expiration_seconds
    except (KeyError, ValueError, StopIteration) as e:
        raise web.HTTPBadRequest(body=f'Invalid template: {e}') from e

async def _get_project_move_template(request: web.Request) -> web.Response:
    logger = logging.getLogger(__name__)
    id_ = request.match_info['id']
    try:
        id = encode_key(split(awsservicelib.decode_folder(id_))[0])
        add_run_time_action(request, name='heaserver-awss3folders-project-get-target', rel='headata-target', path=(
            '/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/' + id) if id else '/volumes/{volume_id}/buckets/{bucket_id}')
        return await _get_project(request)
    except KeyDecodeException as e:
        logger.exception('Error getting parent key')
        return response.status_not_found()

async def _get_project(request: web.Request) -> web.Response:
    """
    Gets the project with the specified id.

    :param request: the HTTP request.
    :return: the requested project or Not Found.
    """
    try:
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        project_id = request.match_info['id'] if 'id' in request.match_info else request.match_info['name']
    except KeyError as e:
        raise ValueError(str(e))
    try:
        project_key = awsservicelib.decode_folder(project_id)
    except KeyDecodeException as e:
        return response.status_not_found()

    async with mongo.MongoContext(request=request, volume_id=volume_id) as mongo_client:
        async with DesktopObjectActionLifecycle(request=request,
                                                code='hea-get',
                                                description=f'Accessing {awsservicelib._activity_object_display_name(bucket_name, project_key)}',
                                                activity_cb=publish_desktop_object) as activity:
            async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3_client:
                try:
                    if not await _is_project(request, volume_id, bucket_name, project_id):
                        return await response.get(request, None)
                    project = await _get_project_helper(s3_client, request)
                    if project is None:
                        return await response.get(request, None)
                    return await response.get(request, project.to_dict())
                except BotoClientError as e:
                    activity.status = Status.FAILED
                    return awsservicelib.handle_client_error(e)
                except web.HTTPClientError as e:
                    activity.status = Status.FAILED
                    return e

async def _get_project_helper(s3_client: S3Client, request: web.Request) -> AWSS3Project:
    try:
        volume_id = request.match_info['volume_id']
        bucket_name = request.match_info['bucket_id']
        project_id = request.match_info['id'] if 'id' in request.match_info else request.match_info['name']
    except KeyError as e:
        raise ValueError(str(e))

    try:
        project_key: str | None = awsservicelib.decode_folder(project_id)
        if not awsservicelib.is_folder(project_key):
            project_key = None
    except KeyDecodeException:
        # Let the bucket query happen so that we consistently return Forbidden if the user lacks permissions
        # for the bucket.
        project_key = None
    if project_key is None:
        # We couldn't decode the folder_id, and we need to check if the user can access the bucket in order to
        # decide which HTTP status code to respond with (Forbidden vs Not Found).
        s3_client.head_bucket(Bucket=bucket_name)
        raise web.HTTPNotFound
    response_ = s3_client.list_objects_v2(Bucket=bucket_name, Prefix=project_key, MaxKeys=1)
    logging.debug('Result of getting project: %s', response_)
    if project_key is None or response_['KeyCount'] == 0:
        raise web.HTTPNotFound
    contents = response_['Contents'][0]
    encoded_key = encode_key(response_['Prefix'])
    return _new_project(bucket_name, contents, encoded_key, request)

async def _is_project(request: web.Request, volume_id: str, bucket_id: str, encoded_key: str):
    async with mongo.MongoContext(request=request, volume_id=volume_id) as mongo_client:
        metadata_dict = await mongo_client.get_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_id, 'encoded_key': encoded_key})
        if not metadata_dict or not issubclass(desktop_object_type_or_type_name_to_type(metadata_dict['actual_object_type_name']), AWSS3Project):
            return False
    return True

def _new_project(bucket_name: str, contents: dict[str, Any], encoded_key: str,
                request: web.Request) -> AWSS3Project:
    project = AWSS3Project()
    project.id = encoded_key
    project.modified = contents['LastModified']
    project.created = contents['LastModified']
    project.owner = request.headers.get(SUB, NONE_USER)
    project.bucket_id = bucket_name
    project.source = AWS_S3
    return project

async def _get_project_move_template(request: web.Request) -> web.Response:
    logger = logging.getLogger(__name__)
    id_ = request.match_info['id']
    try:
        id = encode_key(split(awsservicelib.decode_folder(id_))[0])
        add_run_time_action(request, name='heaserver-awss3folders-project-get-target', rel='headata-target', path=(
            '/volumes/{volume_id}/buckets/{bucket_id}/awss3projects/' + id) if id else '/volumes/{volume_id}/buckets/{bucket_id}')
        return await _get_project(request)
    except KeyDecodeException as e:
        logger.exception('Error getting parent key')
        return response.status_not_found()

async def _move(s3_client: S3Client, source_bucket_id: str, source_key: str, target_bucket_id: str, target_key: str):
    logger = logging.getLogger(__name__)
    with ThreadPoolExecutor(10) as pool:
        tasks = []
        loop = asyncio.get_running_loop()
        def copy_and_delete(source_key, target_key):
            s3_client.copy_object(Bucket=target_bucket_id, CopySource=f'{source_bucket_id}/{source_key}', Key=target_key)
            s3_client.delete_object(Bucket=source_bucket_id, Key=source_key)
        async for object in awsservicelib.list_objects(s3_client, source_bucket_id, source_key, max_keys=1000, loop=loop):
            source_key_ = object['Key']
            target_key_ = replace_parent_folder(source_key=source_key_, target_key=target_key, source_key_folder=source_key)
            logger.debug('Moving %s/%s to %s/%s', source_bucket_id, source_key_, target_bucket_id, target_key_)
            tasks.append(loop.run_in_executor(pool, partial(copy_and_delete, source_key_, target_key_)))
            if len(tasks) % 10 == 0:
                await asyncio.gather(*tasks)
                tasks.clear()
        if tasks:
            await asyncio.gather(*tasks)

async def _get_items_in_container(request: web.Request, volume_id: str, bucket_id: str, folder_id: str) -> web.Response:
    logger = logging.getLogger(__name__)
    async with aws.S3ClientContext(request=request, volume_id=volume_id) as s3:
        folder_key = awsservicelib.decode_folder(folder_id)
        loop = asyncio.get_running_loop()
        try:
            if folder_key is None:
                # We couldn't decode the folder_id, and we need to check if the user can access the bucket in order to
                # decide which HTTP status code to respond with (Forbidden vs Not Found).
                return await awsservicelib.return_bucket_status_or_not_found(bucket_id, loop, s3)
            logger.debug('Getting all items in project %s in bucket %s', folder_key, bucket_id)
            folders: dict[str, AWSS3ItemInFolder] = {}
            actual_object_uri_base = URL('/volumes') / volume_id / 'buckets' / bucket_id
            owner = request.headers.get(SUB, NONE_USER)
            file_type_name = AWSS3FileObject.get_type_name()
        except BotoClientError as e:
            return awsservicelib.handle_client_error(e)
        except ParamValidationError as e:
            return response.status_bad_request(str(e))
        async with DesktopObjectActionLifecycle(request,
                                                code='hea-get',
                                                description=f'Listing all items in {awsservicelib._activity_object_display_name(bucket_id, folder_key)}',
                                                activity_cb=publish_desktop_object) as activity:
            async with mongo.MongoContext(request=request, volume_id=volume_id) as mongo_client:
                try:
                    async for obj in awsservicelib.list_objects(s3, bucket_id=bucket_id, prefix=folder_key, loop=loop):
                        obj_last_modified = obj.get('LastModified')
                        key = obj['Key']
                        key_ = key.removeprefix(folder_key)
                        try:
                            if key_ == '':  # The folder
                                continue
                            actual_key = key_[:key_.index('/') + 1]  # A folder
                            is_folder_ = True
                        except ValueError:
                            actual_key = key_  # Not a folder
                            is_folder_ = False
                        if (item_ := folders.get(actual_key)) is not None:
                            if obj_last_modified is not None:
                                if item_.modified is not None:
                                    item_.modified = max(item_.modified, obj_last_modified)
                                else:
                                    item_.modified = obj_last_modified
                        else:
                            actual_id = encode_key(folder_key + actual_key)
                            logger.debug('Found item %s in bucket %s', actual_key, bucket_id)
                            item = AWSS3ItemInFolder()
                            item.id = actual_id
                            item.modified = obj_last_modified
                            item.created = obj_last_modified
                            item.owner = owner
                            item.actual_object_id = actual_id
                            item.folder_id = folder_id
                            item.storage_class = S3StorageClass[obj['StorageClass']] if not is_folder_ else None
                            item.bucket_id = bucket_id
                            item.source = AWS_S3
                            item.volume_id = volume_id
                            if not is_folder_:
                                item.actual_object_uri = str(actual_object_uri_base / 'awss3files' / actual_id)
                                item.actual_object_type_name = file_type_name
                                item.size = obj['Size']
                            folders[actual_id] = item
                    item_metadata_dict = {}
                    async for metadata_dict in mongo_client.get_all_admin(MONGODB_AWS_S3_FOLDER_METADATA_COLLECTION, {'bucket_id': bucket_id}):
                        item_metadata_dict[metadata_dict['encoded_key']] = metadata_dict
                    actual_folders = []
                    for actual_id, folder in folders.items():
                        if folder.actual_object_type_name != file_type_name:
                            item_metadata = item_metadata_dict.get(folder.id)
                            if item_metadata:
                                aotn = item_metadata.get('actual_object_type_name')
                            else:
                                aotn = None
                            if aotn is None or aotn == AWSS3Folder.get_type_name():
                                folder.actual_object_uri = str(actual_object_uri_base / 'awss3folders' / actual_id)
                                folder.actual_object_type_name = AWSS3Folder.get_type_name()
                            elif aotn == AWSS3Project.get_type_name():
                                folder.actual_object_type_name = AWSS3Project.get_type_name()
                                folder.actual_object_uri = str(actual_object_uri_base / 'awss3projects' / actual_id)
                        actual_folders.append(folder)
                    return await response.get_all(request, list(i.to_dict() for i in actual_folders))
                except BotoClientError as e:
                    activity.status = Status.FAILED
                    return awsservicelib.handle_client_error(e)
                except ParamValidationError as e:
                    activity.status = Status.FAILED
                    return response.status_bad_request(str(e))

def _check_folder_and_object_keys(folder_id_: str | None, request: web.Request) -> tuple[str | None, str | None, bool]:
    folder_or_item_not_found = False
    try:
        decoded_folder_key = decode_key(folder_id_)
    except KeyDecodeException as e:
        folder_or_item_not_found = True
    if decoded_folder_key is None:
        folder_or_item_not_found = True
    try:
        decoded_key = decode_key(request.match_info['id'])
        if not is_object_in_folder(decoded_key, decoded_folder_key):
            folder_or_item_not_found = True
    except KeyDecodeException as e:
        decoded_key = None
        folder_or_item_not_found = True

    return decoded_folder_key, decoded_key, folder_or_item_not_found
